/*
 * Copyright (c) 2007-2018, Douglas Gilbert
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#define __STDC_FORMAT_MACROS 1
#include <inttypes.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "smp_lib.h"
#include "sg_unaligned.h"
#include "sg_pr2serr.h"

/* This is a Serial Attached SCSI (SAS) Serial Management Protocol (SMP)
 * utility.
 *
 * This utility issues a REPORT EXPANDER ROUTE TABLE LIST request and outputs
 * its response.
 */

static const char * version_str = "1.16 20180725";


static struct option long_options[] = {
    {"brief", no_argument, 0, 'b'},
    {"help", no_argument, 0, 'h'},
    {"hex", no_argument, 0, 'H'},
    {"index", required_argument, 0, 'i'},
    {"interface", required_argument, 0, 'I'},
    {"num", required_argument, 0, 'n'},
    {"phy", required_argument, 0, 'p'},
    {"sa", required_argument, 0, 's'},
    {"raw", no_argument, 0, 'r'},
    {"verbose", no_argument, 0, 'v'},
    {"version", no_argument, 0, 'V'},
    {0, 0, 0, 0},
};

struct opts_t {
    bool do_brief;
    bool do_raw;
    int do_hex;
    int start_rsa_index;
    int do_num;
    int phy_id;
    int verbose;
    uint64_t sa;
};


static void
usage(void)
{
    pr2serr("Usage: smp_rep_exp_route_tbl  [--brief] [--help] [--hex] "
            "[--index=IN]\n"
            "                    [--interface=PARAMS] [--num=NUM] [--phy=ID] "
            "[--raw]\n"
            "                    [--sa=SAS_ADDR] "
           );
    pr2serr("[--verbose] [--version]\n"
            "                    <smp_device>[,<n>]\n"
            "  where:\n"
            "    --brief|-b           brief: abridge output\n"
            "    --help|-h            print out usage message\n"
            "    --hex|-H             print response in hexadecimal\n"
            "    --index=IN|-i IN     starting routed SAS address index "
            "(def: 0)\n"
            "    --interface=PARAMS|-I PARAMS    specify or override "
            "interface\n"
            "    --num=NUM|-n NUM     maximum number of descriptors to fetch "
            "(def: 62)\n"
            "    --phy=ID|-p ID       starting phy identifier within bitmap "
            "(def: 0)\n"
            "                         [should be a multiple of 48]\n"
            "    --raw|-r             output response in binary\n"
            "    --sa=SAS_ADDR|-s SAS_ADDR    SAS address of SMP "
            "target (use leading\n"
            "                                 '0x' or trailing 'h'). "
            "Depending on\n"
            "                                 the interface, may not be "
            "needed\n"
           );
    pr2serr("    --verbose|-v         increase verbosity\n"
            "    --version|-V         print version string and exit\n\n"
            "Performs a SMP REPORT EXPANDER ROUTE TABLE LIST function. "
            "Each descriptor\nin the output contains: a routed SAS address, "
            "a 48 bit phy bitmap and a\nzone group\n"
           );
}

static void
dStrRaw(const uint8_t * str, int len)
{
    int k;

    for (k = 0 ; k < len; ++k)
        printf("%c", str[k]);
}

/* Returns 0 when successful, -1 for low level errors and > 0
   for other error categories. */
static int
do_rep_exp_rou_tbl(struct smp_target_obj * top, uint8_t * resp,
                   int max_resp_len, struct opts_t * optsp)
{
    int len, res, k, dword_resp_len, act_resplen;
    char * cp;
    uint8_t smp_req[] = {SMP_FRAME_TYPE_REQ, SMP_FN_REPORT_EXP_ROUTE_TBL_LIST,
                         0, 6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                         0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, };
    char b[256];
    struct smp_req_resp smp_rr;

    dword_resp_len = (max_resp_len - 8) / 4;
    smp_req[2] = (dword_resp_len < 0x100) ? dword_resp_len : 0xff;
    sg_put_unaligned_be16(optsp->do_num, smp_req + 8);
    sg_put_unaligned_be16(optsp->start_rsa_index, smp_req + 10);
    smp_req[19] = optsp->phy_id;
    if (optsp->verbose) {
        pr2serr("    Report expander route table request: ");
        for (k = 0; k < (int)sizeof(smp_req); ++k) {
            if (0 == (k % 16))
                pr2serr("\n      ");
            else if (0 == (k % 8))
                pr2serr(" ");
            pr2serr("%02x ", smp_req[k]);
        }
        pr2serr("\n");
    }
    memset(&smp_rr, 0, sizeof(smp_rr));
    smp_rr.request_len = sizeof(smp_req);
    smp_rr.request = smp_req;
    smp_rr.max_response_len = max_resp_len;
    smp_rr.response = resp;
    res = smp_send_req(top, &smp_rr, optsp->verbose);

    if (res) {
        pr2serr("smp_send_req failed, res=%d\n", res);
        if (0 == optsp->verbose)
            pr2serr("    try adding '-v' option for more debug\n");
        return -1;
    }
    if (smp_rr.transport_err) {
        pr2serr("smp_send_req transport_error=%d\n", smp_rr.transport_err);
        return -1;
    }
    act_resplen = smp_rr.act_response_len;
    if ((act_resplen >= 0) && (act_resplen < 4)) {
        pr2serr("response too short, len=%d\n", act_resplen);
        return SMP_LIB_CAT_MALFORMED;
    }
    len = resp[3];
    if ((0 == len) && (0 == resp[2])) {
        len = smp_get_func_def_resp_len(resp[1]);
        if (len < 0) {
            len = 0;
            if (optsp->verbose > 0)
                pr2serr("unable to determine response length\n");
        }
    }
    len = 4 + (len * 4);        /* length in bytes, excluding 4 byte CRC */
    if ((act_resplen >= 0) && (len > act_resplen)) {
        if (optsp->verbose)
            pr2serr("actual response length [%d] less than deduced length "
                    "[%d]\n", act_resplen, len);
        len = act_resplen;
    }
    if (optsp->do_hex || optsp->do_raw) {
        if (optsp->do_hex)
            hex2stdout(resp, len, 1);
        else
            dStrRaw(resp, len);
        if (SMP_FRAME_TYPE_RESP != resp[0])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[1] != smp_req[1])
            return SMP_LIB_CAT_MALFORMED;
        if (resp[2]) {
            if (optsp->verbose)
                pr2serr("Report expander route table result: %s\n",
                        smp_get_func_res_str(resp[2], sizeof(b), b));
            return resp[2];
        }
        return 0;
    }
    if (SMP_FRAME_TYPE_RESP != resp[0]) {
        pr2serr("expected SMP frame response type, got=0x%x\n", resp[0]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[1] != smp_req[1]) {
        pr2serr("Expected function code=0x%x, got=0x%x\n", smp_req[1],
                resp[1]);
        return SMP_LIB_CAT_MALFORMED;
    }
    if (resp[2]) {
        cp = smp_get_func_res_str(resp[2], sizeof(b), b);
        pr2serr("Report expander route table result: %s\n", cp);
        return resp[2];
    }
    return 0;
}


int
main(int argc, char * argv[])
{
    int res, c, len, exp_cc, sphy_id, num_desc, desc_len;
    int k, j, off, exp_rtcc;
    int ret = 0;
    int subvalue = 0;
    int64_t sa_ll;
    char * cp;
    char i_params[256];
    char device_name[512];
    uint8_t resp[1020 + 8];
    struct smp_target_obj tobj;
    struct opts_t opts;

    memset(&opts, 0, sizeof(opts));
    opts.do_num = 62;   /* maximum fitting in one response */
    memset(device_name, 0, sizeof device_name);
    while (1) {
        int option_index = 0;

        c = getopt_long(argc, argv, "bhHi:I:n:p:rs:vV", long_options,
                        &option_index);
        if (c == -1)
            break;

        switch (c) {
        case 'b':
            opts.do_brief = true;
            break;
        case 'h':
        case '?':
            usage();
            return 0;
        case 'H':
            ++opts.do_hex;
            break;
        case 'i':
           opts.start_rsa_index = smp_get_num(optarg);
           if ((opts.start_rsa_index < 0) || (opts.start_rsa_index > 65535)) {
                pr2serr("bad argument to '--index'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            break;
        case 'I':
            strncpy(i_params, optarg, sizeof(i_params));
            i_params[sizeof(i_params) - 1] = '\0';
            break;
        case 'n':
           opts.do_num = smp_get_num(optarg);
           if ((opts.do_num < 0) || (opts.do_num > 65535)) {
                pr2serr("bad argument to '--num'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            break;
        case 'p':
           opts.phy_id = smp_get_num(optarg);
           if ((opts.phy_id < 0) || (opts.phy_id > 254)) {
                pr2serr("bad argument to '--phy', expect value from 0 to "
                        "254\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            break;
        case 'r':
            opts.do_raw = true;
            break;
        case 's':
           sa_ll = smp_get_llnum_nomult(optarg);
           if (-1LL == sa_ll) {
                pr2serr("bad argument to '--sa'\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
            opts.sa = (uint64_t)sa_ll;
            break;
        case 'v':
            ++opts.verbose;
            break;
        case 'V':
            pr2serr("version: %s\n", version_str);
            return 0;
        default:
            pr2serr("unrecognised switch code %c [0x%x] ??\n", c, c);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (optind < argc) {
        if ('\0' == device_name[0]) {
            strncpy(device_name, argv[optind], sizeof(device_name) - 1);
            device_name[sizeof(device_name) - 1] = '\0';
            ++optind;
        }
        if (optind < argc) {
            for (; optind < argc; ++optind)
                pr2serr("Unexpected extra argument: %s\n", argv[optind]);
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == device_name[0]) {
        cp = getenv("SMP_UTILS_DEVICE");
        if (cp)
            strncpy(device_name, cp, sizeof(device_name) - 1);
        else {
            pr2serr("missing device name on command line\n    [Could use "
                    "environment variable SMP_UTILS_DEVICE instead]\n\n");
            usage();
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if ((cp = strchr(device_name, SMP_SUBVALUE_SEPARATOR))) {
        *cp = '\0';
        if (1 != sscanf(cp + 1, "%d", &subvalue)) {
            pr2serr("expected number after separator in SMP_DEVICE name\n");
            return SMP_LIB_SYNTAX_ERROR;
        }
    }
    if (0 == opts.sa) {
        cp = getenv("SMP_UTILS_SAS_ADDR");
        if (cp) {
           sa_ll = smp_get_llnum_nomult(cp);
           if (-1LL == sa_ll) {
                pr2serr("bad value in environment variable "
                        "SMP_UTILS_SAS_ADDR\n    use 0\n");
                sa_ll = 0;
            }
            opts.sa = (uint64_t)sa_ll;
        }
    }
    if (opts.sa > 0) {
        if (! smp_is_naa5(opts.sa)) {
            pr2serr("SAS (target) address not in naa-5 format (may need "
                    "leading '0x')\n");
            if ('\0' == i_params[0]) {
                pr2serr("    use '--interface=' to override\n");
                return SMP_LIB_SYNTAX_ERROR;
            }
        }
    }

    res = smp_initiator_open(device_name, subvalue, i_params, opts.sa,
                             &tobj, opts.verbose);
    if (res < 0)
        return SMP_LIB_FILE_ERROR;

    ret = do_rep_exp_rou_tbl(&tobj, resp, sizeof(resp), &opts);
    if (ret)
        goto finish;
    if (opts.do_hex || opts.do_raw)
        goto finish;
    len = (resp[3] * 4) + 4;    /* length in bytes excluding CRC field */
    exp_cc = sg_get_unaligned_be16(resp + 4);
    exp_rtcc = sg_get_unaligned_be16(resp + 6);
    desc_len = resp[10];
    num_desc = resp[11];
    sphy_id = resp[19];
    printf("Report expander route table response header:\n");
    if (! opts.do_brief) {
        printf("  expander change count: %d\n", exp_cc);
        printf("  expander route table change count: %d\n", exp_rtcc);
        printf("  self configuring: %d\n", !!(resp[8] & 0x8));
        printf("  zone configuring: %d\n", !!(resp[8] & 0x4));
        printf("  configuring: %d\n", !!(resp[8] & 0x2));
        printf("  zone enabled: %d\n", !!(resp[8] & 0x1));
        printf("  expander route table descriptor length: %d dwords\n",
               desc_len);
    }
    printf("  number of expander route table descriptors: %d\n", num_desc);
    printf("  first routed SAS address index: %d\n",
           sg_get_unaligned_be16(resp + 12));
    printf("  last routed SAS address index: %d\n",
           sg_get_unaligned_be16(resp + 14));
    printf("  starting phy id: %d\n", sphy_id);

    if (len != (32 + (num_desc * (desc_len * 4)))) {
        pr2serr(">>> Response length of %d bytes doesn't match %d "
                "descriptors, each\n  of %d bytes plus a 32 byte header and "
                "4 byte CRC\n", len + 4, num_desc, desc_len * 4);
        if (len < (32 + (num_desc * (desc_len * 4)))) {
            ret = SMP_LIB_CAT_MALFORMED;
            goto finish;
        }
    }
    for (k = 0; k < num_desc; ++k) {
        off = 32 + (k * (desc_len * 4));
        printf("  descriptor index %d:\n", k);
        printf("    routed SAS address: 0x%" PRIx64 "\n",
               sg_get_unaligned_be64(resp + off));
        printf("    phy bit map: 0x");
        for (j = 0; j < 6; ++j)
            printf("%02x", resp[off + 8 + j]);
        printf("\n");
        printf("    zone group: %d\n", resp[off + 15]);
    }

finish:
    res = smp_initiator_close(&tobj);
    if (res < 0) {
        if (0 == ret)
            return SMP_LIB_FILE_ERROR;
    }
    if (ret < 0)
        ret = SMP_LIB_CAT_OTHER;
    if (opts.verbose && ret)
        pr2serr("Exit status %d indicates error detected\n", ret);
    return ret;
}
